from copy import deepcopy
from enum import auto, Enum
from typing import List

from energyplus_pet.correction_factor import CorrectionFactor, CorrectionFactorType


class CatalogDataManager:
    """
    This class represents a data manager for the entire catalog data set.  This includes
    the primary tabular data, plus any correction factors applied to the data.
    While the equipment definitions define the types of data, the data manager actually stores the data.
    The equipment instances read data from the catalog data manager when processing parameters.
    """

    def __init__(self):
        """
        Create a new CatalogDataManager instance, initializing arrays and flags.
        """
        self._correction_factors: List[CorrectionFactor] = []
        self._base_data: List[List[float]] = []  # inner arrays are column allocated: self.base_data[data_point][column]
        self.data_processed = False
        self.final_data_matrix: List[List[float]] = []
        self.last_error_message = ""

    def add_correction_factor(self, cf: CorrectionFactor) -> None:
        """
        Add a completed correction factor, with summary data and detailed data.

        :param cf: A correction factor instance, expecting to be fully fleshed out with summary and detailed data.
        :return: None
        """
        self._correction_factors.append(cf)

    # TODO: I think a nicer interface here would be add_base_data_column(column_id), and each equip defines column ids
    def add_base_data(self, data: List[List[float]]) -> None:
        """
        Add base data as a list of data point rows, so the array lookup should be data[row][column].
        The data should already be in the calculation_unit for each column.

        Right now this function does not do any checks on the data because the tabular forms are supposed to handle
        all of that, and the apply_correction_factors function does a bunch of checking as well.

        :param data: Catalog base data set in proper units
        :return: None
        """
        self._base_data = data

    def summary(self) -> dict:
        """Returns a string representation of the catalog data manager as it currently exists"""
        return {
            'base_data_in_rows': self._base_data,
            'correction_factors': [cf.describe() for cf in self._correction_factors],
            'final_data_rows': self.final_data_matrix
        }

    class ProcessResult(Enum):
        OK = auto()
        ERROR = auto()

    def apply_correction_factors(self, minimum_data_points: int, db_column: int, wb_column: int) -> ProcessResult:
        """
        Process the base data and correction factors to create one large full dataset.
        Validates the data against a series of tests for data diversity and infinite/out-of-range.

        :return: A ProcessResult enum instance for the success of the process.  If ERROR, then there is a
                 ``last_error_message`` member variable with an explanation of what went wrong.
        """
        self.data_processed = True
        self.final_data_matrix = self._base_data
        for cf in self._correction_factors:
            updated_data_matrix = deepcopy(self.final_data_matrix)  # deep is required for complex lists of lists
            for cf_row in range(cf.num_corrections):  # each row of the cf data implies a new copy of the data set
                for row in updated_data_matrix:
                    new_row = list(row)  # list provides a deep copy of a simple list
                    if cf.correction_type == CorrectionFactorType.Multiplier:
                        new_row[cf.base_column_index] *= cf.base_correction[cf_row]
                    elif cf.correction_type == CorrectionFactorType.Replacement:
                        new_row[cf.base_column_index] = cf.base_correction[cf_row]
                    elif cf.correction_type == CorrectionFactorType.CombinedDbWb:
                        new_row[db_column] = cf.base_correction_db[cf_row]
                        new_row[wb_column] = cf.base_correction_wb[cf_row]
                    for column_to_modify in cf.columns_to_modify:
                        new_row[column_to_modify] *= cf.mod_correction_data_column_map[column_to_modify][cf_row]
                    self.final_data_matrix.append(new_row)
        if len(self.final_data_matrix) < minimum_data_points:
            self.last_error_message = f"Full catalog data set too small. \nData includes {len(self.final_data_matrix)} "
            self.last_error_message += f"rows, but this equipment requires at least {minimum_data_points}."
            return CatalogDataManager.ProcessResult.ERROR
        else:
            if len(self.final_data_matrix) == 0:
                self.last_error_message = "Catalog data appears empty!  Abort!"
                return CatalogDataManager.ProcessResult.ERROR
            for column_index in range(len(self.final_data_matrix[0])):
                this_column_data = [row[column_index] for row in self.final_data_matrix]
                if len(set(this_column_data)) == 1:
                    self.last_error_message = f"Problem with data, column #{column_index} (zero-based) is constant "
                    self.last_error_message += "after factors have been applied.  Each column should contain variation!"
                    return CatalogDataManager.ProcessResult.ERROR
        return CatalogDataManager.ProcessResult.OK

    def reset(self) -> None:
        """
        Resets the catalog data manager to an original state.

        :return: None
        """
        self._correction_factors.clear()
        self.data_processed = False
        self._base_data = []
        self.final_data_matrix: List[List[float]] = []
        self.last_error_message = ""



from typing import Type, Union

from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.equipment.equip_types import EquipType
from energyplus_pet.equipment.wahp_heating_curve import WaterToAirHeatPumpHeatingCurveFit
from energyplus_pet.equipment.wahp_cooling_curve import WaterToAirHeatPumpCoolingCurveFit
from energyplus_pet.equipment.wwhp_heating_curve import WaterToWaterHeatPumpHeatingCurveFit
from energyplus_pet.equipment.wwhp_cooling_curve import WaterToWaterHeatPumpCoolingCurveFit


class EquipmentFactory:
    """Handles construction of equipment"""

    @staticmethod
    def class_factory(equipment_type: EquipType) -> Union[Type[BaseEquipment], None]:
        """Returns a class Type for the given EquipType enum value, or None if no match"""
        type_map = {
            EquipType.InvalidType: None,
            EquipType.WAHP_Heating_CurveFit: WaterToAirHeatPumpHeatingCurveFit,
            EquipType.WAHP_Heating_PE: None,
            EquipType.WAHP_Cooling_CurveFit: WaterToAirHeatPumpCoolingCurveFit,
            EquipType.WAHP_Cooling_PE: None,
            EquipType.WWHP_Heating_CurveFit: WaterToWaterHeatPumpHeatingCurveFit,
            EquipType.WWHP_Cooling_CurveFit: WaterToWaterHeatPumpCoolingCurveFit,
            EquipType.Pump_ConstSpeed_ND: None,
        }
        return type_map.get(equipment_type, None)

    @staticmethod
    def instance_factory(equipment_type: EquipType) -> Union[BaseEquipment, None]:
        """Returns a class instance for the given EquipType enum value, or None if no match"""
        equip_type = EquipmentFactory.class_factory(equipment_type)
        if equip_type is None:
            return None
        return equip_type()



from enum import auto, Enum

import numpy as np
from tkinter import Frame, Toplevel, TOP, BOTH, Tk, Label, X, Button
from tkinter.ttk import Notebook, Separator

from energyplus_pet.data_manager import CatalogDataManager
from energyplus_pet.equipment.base import BaseEquipment
from energyplus_pet.exceptions import EnergyPlusPetException

import matplotlib
matplotlib.use('TkAgg')

# unfortunately, with newer flake8, you cannot tag the .use() line with noqa, you have to tag *every* trailing import
# I considered just ignoring E402 for this project, as the files are generally small anyway, but for now, noqa here

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg  # noqa: E402
from matplotlib.figure import Figure  # noqa: E402


class ComparisonPlot(Toplevel):
    """
    This form is where we show the user the numeric results of the parameter generation process
    """

    class PlotType(Enum):
        RawComparison = auto()
        PercentError = auto()

    def __init__(self, parent_window: Tk, _: CatalogDataManager, equip_instance: BaseEquipment):
        super().__init__(parent_window)
        self.title(f"{parent_window.title()}: Results Comparison")
        p = 4
        data_per_type = {
            ComparisonPlot.PlotType.RawComparison: {
                "tab_title": "Catalog Data Raw Comparison",
                "plot_title": "Model vs. Catalog Data Points",
                "plot_data": equip_instance.get_absolute_plot_data(),
                "y_label": "[Data could be multiple units]",
            },
            ComparisonPlot.PlotType.PercentError: {
                "tab_title": "Percent Error Comparison",
                "plot_title": "Percent Error (hopefully near zero!)",
                "plot_data": equip_instance.get_error_plot_data(),
                "y_label": "Percent Error [%]",
            }
        }
        plot_notebook = Notebook(self)
        for plot_type in list(ComparisonPlot.PlotType):
            data_this_plot_type = data_per_type[plot_type]
            plot_frame = Frame(plot_notebook)
            plot_frame.pack(side=TOP, expand=True, fill=BOTH)
            plot_notebook.add(plot_frame, text=data_this_plot_type['tab_title'])
            fig = Figure()
            a = fig.add_subplot(111)
            plot_title = data_this_plot_type['plot_title']
            for pd in data_this_plot_type['plot_data']:
                if pd[1] == 'line':
                    line_arg = '-'
                elif pd[1] == 'point':
                    line_arg = '--'
                else:
                    raise EnergyPlusPetException('bad line type')
                a.plot(np.array(pd[3]), label=pd[0], linestyle=line_arg, color=pd[2])
            a.legend()
            a.set_title(plot_title, fontsize=16)
            a.set_xlabel("Catalog Data Points (no order)")
            a.set_ylabel(data_this_plot_type['y_label'])
            canvas = FigureCanvasTkAgg(fig, master=plot_frame)
            canvas.get_tk_widget().pack(side=TOP, expand=True, fill=BOTH)
            canvas.draw()
        plot_notebook.pack(side=TOP, expand=True, fill=BOTH, padx=p, pady=p)
        metrics = [f"{m[0]}: {m[1]}"for m in equip_instance.get_extra_regression_metrics()]
        if metrics:
            Label(self, text='\n'.join(metrics)).pack(side=TOP, expand=False, fill=X)
        Separator(self, orient='horizontal').pack(fill=X, padx=p, pady=p)
        Button(self, text="Wizard Complete!  Click here or close the window to finish", command=self.close_me).pack(
            side=TOP, expand=False, padx=p, pady=p
        )
        self.grab_set()
        self.transient(parent_window)

    def close_me(self):
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_heating_curve import WaterToAirHeatPumpHeatingCurveFit
    window = Tk()
    cdm = CatalogDataManager()
    eq = WaterToAirHeatPumpHeatingCurveFit()
    eq.catalog_heating_capacity = [100.0, 200.0, 300.0]
    eq.predicted_heating_capacity = [100.00001, 200.00001, 299.99999]
    eq.percent_error_heating_capacity = [0.0000001, 0.00000001, 0.0000001]
    eq.heating_capacity_avg_err = 0.00005
    ComparisonPlot(window, cdm, eq)
    window.mainloop()



import pathlib
from setuptools import setup

from energyplus_pet import VERSION

readme_file = pathlib.Path(__file__).parent.resolve() / 'README.md'
readme_contents = readme_file.read_text()

setup(
    name="energyplus_pet",
    version=VERSION,
    packages=['energyplus_pet', 'energyplus_pet.forms', 'energyplus_pet.equipment'],
    description="Parameter Estimation Tools for Generating EnergyPlus Inputs from Raw Performance Data",
    package_data={"energyplus_pet.forms": ["*.png", "*.ico"], "energyplus_pet": ["examples/*.ods"]},
    include_package_data=True,
    long_description=readme_contents,
    long_description_content_type='text/markdown',
    author='Edwin Lee',
    author_email='a@a.a',
    url='https://github.com/Myoldmopar/EnergyPlusPet',
    license='UnlicensedForNow',
    install_requires=['pyperclip', 'tksheet', 'matplotlib', 'numpy', 'scipy', 'pyshortcuts'],
    entry_points={
        'console_scripts': ['energyplus_pet_gui=energyplus_pet.runner:main_gui']
    }
)



from enum import Enum, auto
from functools import partial
from platform import system
from tkinter import Toplevel, Frame, Canvas, simpledialog
from tkinter import Button, Label, Listbox, Scrollbar
from tkinter import HORIZONTAL, TOP, X, BOTH, VERTICAL, NW, LEFT, Y, EW, RIGHT
from tkinter import StringVar
from tkinter.ttk import Separator
from typing import List

from energyplus_pet.correction_factor import CorrectionFactor
from energyplus_pet.forms.correction_summary_widget import CorrectionSummaryWidget
from energyplus_pet.forms.basic_message_form import PetMessageForm
from energyplus_pet.equipment.base import BaseEquipment


class Event:
    LinuxWheelUp = '<Button-4>'
    LinuxWheelDown = '<Button-5>'
    WindowsWheelEvent = '<MouseWheel>'
    Configure = '<Configure>'
    WidgetEnter = '<Enter>'
    WidgetLeave = '<Leave>'


class CorrectionFactorSummaryForm(Toplevel):
    """
    This form is where the user will enter basic correction factor summary data.
    """
    class ExitCode(Enum):
        Cancel = auto()
        Done = auto()
        Error = auto()

    def __init__(self, parent_window, equipment: BaseEquipment):
        super().__init__(parent_window, height=200, width=200)
        self.title(f"{parent_window.title()}: Correction Factor Summary Input")
        # store arguments for manipulation here, these are passed by assignment, in this case "by reference"
        self._summary_widgets: List[CorrectionSummaryWidget] = []
        self.factor_summaries: List[CorrectionFactor] = []
        self._equipment = equipment
        self.exit_code = CorrectionFactorSummaryForm.ExitCode.Cancel  # initialize with this value
        self._text_done = 'Done, ready for factor details'
        self._text_skip = 'Skip entering factors'
        # create the gui
        self._build_gui()
        # draw factors, in case there already are any
        self._redraw_factors()
        # finalize UI operations
        self.grab_set()
        self.transient(parent_window)

    def _build_gui(self):
        lbl = Label(self, text="""All data columns should have at least two values.
Without this variation, a curve fit cannot be generated.
Manufacturers often provide tabular data at fixed temperatures/flow.
However, there is often correction factor data that can add variation.
These factors modify the "constant" temp/flow values by replacement or multiplying.
If you have any correction factors, add them here, otherwise, press done to continue.""")
        s_0 = Separator(self, orient=HORIZONTAL)
        #
        correction_factor_outer_frame = Frame(self)
        # need to store the canvas as a member since it is "scrolled" to view the inner frame
        self._factor_canvas = Canvas(correction_factor_outer_frame)
        # need to store the scrollbar because we check what widget we "scrolled" later
        self._scrollbar = Scrollbar(correction_factor_outer_frame, orient=VERTICAL, command=self._factor_canvas.yview)
        # need to store this because it is the actual frame where we place correction factor summaries
        self._correction_factor_inner_frame = Frame(self._factor_canvas, height=20)
        # bind the inner frame's configure event to a lambda that will update the containing canvas
        self._correction_factor_inner_frame.bind(
            Event.Configure, lambda e: self._factor_canvas.configure(scrollregion=self._factor_canvas.bbox("all"))
        )
        self._correction_factor_inner_frame.grid_columnconfigure(0, weight=1)
        # create a window into the inner frame to display on the canvas
        self._factor_canvas.create_window((0, 0), window=self._correction_factor_inner_frame, anchor=NW)
        # configure the canvas to update the scrollbar when the view is changed in any other way
        self._factor_canvas.configure(yscrollcommand=self._scrollbar.set)
        # pack the canvas and the scrollbar inside the outer frame
        self._factor_canvas.pack(side=LEFT, fill=BOTH, expand=True, padx=3, pady=3)
        self._scrollbar.pack(side=RIGHT, fill=Y, padx=3, pady=3)
        # bind the mouse entry/exit events to register or deregister mouse wheel listeners
        self._factor_canvas.bind(Event.WidgetEnter, self._bind_mouse_wheel_events)
        self._factor_canvas.bind(Event.WidgetLeave, self._unbind_mouse_wheel_events)
        #
        s_1 = Separator(self, orient=HORIZONTAL)
        button_frame = Frame(self)
        btn_add = Button(button_frame, text="Add Correction Factor (Ctrl-a)", command=self._add_factor_widget)
        self._txt_done_skip = StringVar(value=self._text_skip)
        btn_ok_skip = Button(button_frame, textvariable=self._txt_done_skip, command=self._done_skip)
        btn_cancel = Button(button_frame, text="Cancel Wizard", command=self._cancel)
        # pack everything
        lbl.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        s_0.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        correction_factor_outer_frame.pack(side=TOP, fill=BOTH, expand=True, padx=3, pady=3)
        s_1.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        button_frame.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        btn_add.grid(row=0, column=0, padx=3, pady=3)
        btn_ok_skip.grid(row=0, column=1, padx=3, pady=3)
        btn_cancel.grid(row=0, column=2, padx=3, pady=3)
        # configure the grid
        button_frame.grid_columnconfigure(0, weight=1)
        button_frame.grid_columnconfigure(1, weight=1)
        button_frame.grid_columnconfigure(2, weight=1)
        self.bind('<Key>', self._handle_button_pressed)

    def _handle_button_pressed(self, event):
        # relevant_modifiers
        # mod_shift = 0x1
        mod_control = 0x4
        # mod_alt = 0x20000
        if event.keysym == 'a' and mod_control & event.state:
            self._add_factor_widget()

    def _handle_mouse_wheel_event_linux(self, event, scroll):
        amount_to_scroll_canvas = int(scroll)
        if event.widget == self._scrollbar:
            amount_to_scroll_canvas = int(amount_to_scroll_canvas / 2.0)
        elif isinstance(event.widget, Listbox):
            amount_to_scroll_canvas = 0
        self._factor_canvas.yview_scroll(amount_to_scroll_canvas, "units")

    def _handle_mouse_wheel_event_windows(self, event):
        amount_to_scroll_canvas = int(-1 * (event.delta / 120))
        if event.widget == self._scrollbar:
            amount_to_scroll_canvas = int(amount_to_scroll_canvas / 2.0)
        elif isinstance(event.widget, Listbox):
            amount_to_scroll_canvas = 0
        self._factor_canvas.yview_scroll(amount_to_scroll_canvas, "units")

    def _bind_mouse_wheel_events(self, _):
        if system() == 'Linux':
            self._factor_canvas.bind_all(Event.LinuxWheelUp, partial(self._handle_mouse_wheel_event_linux, scroll=-1))
            self._factor_canvas.bind_all(Event.LinuxWheelDown, partial(self._handle_mouse_wheel_event_linux, scroll=1))
        elif system() == 'Windows':
            self._factor_canvas.bind_all(Event.WindowsWheelEvent, self._handle_mouse_wheel_event_windows)

    def _unbind_mouse_wheel_events(self, _):
        if system() == 'Linux':
            self._factor_canvas.unbind_all(Event.LinuxWheelUp)
            self._factor_canvas.unbind_all(Event.LinuxWheelDown)
        elif system() == 'Windows':
            self._factor_canvas.unbind_all(Event.WindowsWheelEvent)

    def _redraw_factors(self):
        for i, f in enumerate(self._summary_widgets):
            f.grid(row=i, column=0, sticky=EW, padx=3, pady=3)

    def _add_factor_widget(self):
        name = simpledialog.askstring(
            "Correction Factor Name", "Give this correction factor a name", initialvalue="CorrFactorName", parent=self
        )
        if name is None:
            return
        name = name.lower()
        new_widget = CorrectionSummaryWidget(
            self._correction_factor_inner_frame, name, self._equipment, self._remove_a_factor
        )
        self._summary_widgets.append(new_widget)
        self._redraw_factors()
        self._txt_done_skip.set(self._text_done)

    def _remove_a_factor(self):
        indexes_to_remove = []
        for i, f in enumerate(self._summary_widgets):
            if f.remove_me:
                indexes_to_remove.append(i)
        for i in reversed(indexes_to_remove):
            self._summary_widgets[i].destroy()
            del self._summary_widgets[i]
        self._redraw_factors()
        if len(self._summary_widgets) == 0:
            self._txt_done_skip.set(self._text_skip)

    def _done_skip(self):
        self.exit_code = CorrectionFactorSummaryForm.ExitCode.Done
        output_message = ""
        for x in self._summary_widgets:
            if not x.check_ok():
                output_message += f"Errors for {x.cf.name}:\n"
                for e in x.cf.check_ok_messages:
                    output_message += f" - {e}\n"
        if output_message:
            message_window = PetMessageForm(
                self, "Problem with Correction Factor(s)", output_message, justify_message_left=True
            )
            self.wait_window(message_window)
            return
        self.factor_summaries = [x.cf for x in self._summary_widgets]
        self.grab_release()
        self.destroy()

    def _cancel(self):
        self.exit_code = CorrectionFactorSummaryForm.ExitCode.Cancel
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wahp_heating_curve import WaterToAirHeatPumpHeatingCurveFit

    root = Tk()
    eq = WaterToAirHeatPumpHeatingCurveFit()
    CorrectionFactorSummaryForm(root, eq)
    root.mainloop()



from tkinter import Toplevel, Button, Frame, Label, HORIZONTAL, TOP, X
from tkinter.ttk import Separator

from pyperclip import copy

from energyplus_pet.equipment.base import BaseEquipment


class RequiredDataPreviewForm(Toplevel):
    """Allows users to view the required equipment data and copy it to the clipboard"""
    def __init__(self, parent_window, equipment: BaseEquipment):
        super().__init__(parent_window)
        self.title(f"{parent_window.title()}: Header Data Summary")
        headers_summary = equipment.headers().get_descriptive_summary().strip()
        headers_csv = equipment.headers().get_descriptive_csv()
        Label(self, text=equipment.name()).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        Label(self, text="Data columns listed here\nData can be entered in any common units").pack(
            side=TOP, fill=X, expand=False, padx=3, pady=3
        )
        Label(self, text=headers_summary).pack(side=TOP, fill=X, expand=True, padx=3, pady=3)
        Label(
            self,
            text="If any data uses correction factors,\n constant values can be entered in tabular form \n and "
                 "correction factors are separate from tabulated data. "
        ).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        Label(self, text="The equipment needs fixed parameters:").pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        constant_parameters = equipment.get_required_constant_parameters()
        param_summary = '\n'.join([f"{c.title} [{c.unit_type}]" for c in constant_parameters])
        param_csv = "ParameterName,UnitType\n"
        param_csv += '\n'.join([f"{c.title},{c.unit_type}" for c in constant_parameters])
        Label(self, text=param_summary).pack(side=TOP, fill=X, expand=True, padx=3, pady=3)
        self._summary = f"{equipment.name()}\n*Columnar data*\n{headers_summary}\n*Parameters*\n{param_summary}"
        self._csv = f"{equipment.short_name()}\n{headers_csv}\n\n{param_csv}"
        Separator(self, orient=HORIZONTAL).pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        button_frame = Frame(self)
        Button(button_frame, text="Copy Summary to Clipboard", command=self._copy).grid(row=0, column=0, padx=3, pady=3)
        Button(button_frame, text="Copy CSV to Clipboard", command=self._copy_csv).grid(row=0, column=1, padx=3, pady=3)
        Button(button_frame, text="Done", command=self._done).grid(row=0, column=2, padx=3, pady=3)
        button_frame.grid_columnconfigure(0, weight=1)
        button_frame.grid_columnconfigure(1, weight=1)
        button_frame.grid_columnconfigure(2, weight=1)
        button_frame.pack(side=TOP, fill=X, expand=False, padx=3, pady=3)
        # finalize UI operations
        self.grab_set()
        self.transient(parent_window)

    def _copy(self):
        copy(self._summary)

    def _copy_csv(self):
        copy(self._csv)

    def _done(self):
        self.grab_release()
        self.destroy()


if __name__ == "__main__":
    from tkinter import Tk
    from energyplus_pet.equipment.wwhp_heating_curve import WaterToWaterHeatPumpHeatingCurveFit
    root = Tk()
    RequiredDataPreviewForm(root, WaterToWaterHeatPumpHeatingCurveFit())
    root.mainloop()
